/*======================================================================
  
  A TOSHIBA LAN Card PCMCIA client driver 
  Please send a bug report at masahiro@isl.rdc.toshiba.co.jp
  
  This is a driver for the Toshiba LAN Interface Cards (LANCT00A)
  under the Linux pcmcia card manager written by 
  David Hinds, dhinds@allegro.stanford.edu

  Written by Masahiro =Rhythm Drive= ISHIYAMA
  Communication Platform Laboratory,
  Research and Development Center, Toshiba Corporation.

  Copyright (C) 1995-1999 Toshiba Corporation.

  Based on the code written 1994 by Hideki YOSHIDA.

  Based on pcnet_cs.c written by Donald Becker.
  Copyright 1993 United States Government as represented by the
  Director, National Security Agency.  This software may be used and
  distributed according to the terms of the GNU Public License,
  incorporated herein by reference.
  Donald Becker may be reached at becker@cesdis1.gsfc.nasa.gov
  
  ======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/system.h>

#include <linux/netdevice.h>
#include <drivers/net/8390.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

#undef DEBUG_FORCE

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
#else
#define DEBUG(n, args...)
#endif

#ifdef DEBUG_FORCE
#undef DEBUG
#define DEBUG(n, args...) printk(KERN_NOTICE args)
#endif

static char *version =
"$Id: tos_cs.c,v 2.18 1999/06/02 08:30:28 masahiro Exp $ (Masahiro =Rhythm Drive= Ishiyama)";


/*====================================================================*/
/*
 * TOSHIBA LAN Card specific params 
 */

#define DS_BASE (dev->base_addr) /* IO base addr of 8390 controller */

#define TOS_START_PG     0x00     /* First page of TX buffer */
#define TOS_STOP_PG      0x80     /* Last page +1 of RX ring */
#define TOS_CM_OFFSET    0

#define TOS_CMDREG       0        /* Offset to ASIC command register. */
#define TOS_CMDREG1      1        /* Offset to ASIC command register1. */
#define TOS_CMDREG2	 2	  /* Offset to ASIC command register1. */
#define TOS_RESET        0x80     /* Board reset, in TOS_CMDREG. */
#define TOS_MEMENB       0x40     /* Enable the shared memory. */
#define TOS_POWERON      0x02     /* Power on. */
#define TOS_NIC_OFFSET   16       /* Offset to the 8390 NIC from the base_addr. */
#define TOS_EADDR_OFFSET 8        /* Offset to the EADDR region from the base */
#define TOS_IO_EXTENT    32       /* IO window size*/
#define TOS_KICKREG_OFFSET 0xc0	  /* magical registar */
#define TOS_KICKREG_VAL    0x50	  /* a spell beyond the wit of man */

/*====================================================================*/

/*====================================================================*/
/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
/* This means pick from 10, 9, 7, 5, 4, and 3 */
static u_long irq_mask = 0x6b8;
/* static int irq_list[4] = { -1 }; */

MODULE_PARM(irq_mask, "i");
/* MODULE_PARM(irq_list, "1-4i"); */

/* Shared memory speed, in ns */
static int mem_speed = 0;
MODULE_PARM(mem_speed, "i");

/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card insertion
   and ejection events.  They are invoked from the tos event
   handler.
*/

static void tos_config(dev_link_t *link);
static void tos_release(u_long arg);
static int tos_event(event_t event, int priority,
			  event_callback_args_t *args);

static int tos_open(struct device *dev);
static int tos_close(struct device *dev);

static void tos_reset_8390(struct device *dev);
static int set_config(struct device *dev, struct ifmap *map);
static int setup_shmem_window(dev_link_t *link, int start_pg,
			      int stop_pg, int cm_offset);

static int tos_setup_config_register(dev_link_t *link);
static int tossetup(dev_link_t *link);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *tos_attach(void);
static void tos_detach(dev_link_t *);

/*
   You'll also need to prototype all the functions that will actually
   be used to talk to your device.  See 'pcmem_cs' for a good example
   of a fully self-sufficient driver; the other drivers rely more or
   less on other parts of the kernel.
*/

/* void tos_interrupt(int reg); */ /* unused */

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "tos_cs";

/*
   A linked list of "instances" of the tos device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.
*/

typedef struct tos_dev_t {
    struct device	dev;
    dev_node_t		node;
    caddr_t             base;
    /* for future use */
} tos_dev_t;

typedef struct hw_info_t {
    char	*id;
    u_long	offset;
    u_char	a0, a1, a2;
    u_long	flags;
} hw_info_t;

static hw_info_t hw_info[] = {
    { "TOSHIBA LAN Card", 0x00, 0x00, 0x00, 0x00, 0 },
};

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

/*======================================================================

    We never need to do anything when a pcnet device is "initialized"
    by the net software, because we only register already-found cards.

======================================================================*/

static int tos_init(struct device *dev)
{
    return 0;
}

/*======================================================================

    tos_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/

static dev_link_t *tos_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    tos_dev_t *info;
    struct device *dev;
    int ret;
    
    DEBUG(0, "tos_attach()\n");

    /* Create new ethernet device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &tos_release;
    link->release.data = (u_long)link;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = irq_mask;
    link->irq.Handler = &ei_interrupt;

    /* The io structure describes IO port mapping */
    link->io.NumPorts1 = 32;
    link->io.NumPorts2 = 0;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.IOAddrLines = 5;

    
    /* General socket configuration */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;

    /* attach ethernet device */
    info = kmalloc(sizeof(struct tos_dev_t), GFP_KERNEL);
    memset(info, 0, sizeof(struct tos_dev_t));
    dev = &info->dev;
    DEBUG(0, "calling ethdev_init()\n");
    ethdev_init(dev);
    dev->name = info->node.dev_name;
    dev->init = &tos_init;
    dev->open = &tos_open;
    dev->stop = &tos_close;
    dev->tbusy = 1;
    dev->set_config = &set_config;
    link->priv = info;
#if	CS_RELEASE_CODE > 0x2911
	link->irq.Instance = dev;
#endif

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &tos_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    DEBUG(0, "Trying CardServices/RegisterClient\n");
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
	cs_error(link->handle, RegisterClient, ret);
	tos_detach(link);
	return NULL;
    }

    return link;
} /* tos_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void tos_detach(dev_link_t *link)
{
    dev_link_t **linkp;
    long flags;

    DEBUG(0, "pcnet_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    save_flags(flags);
    cli();
    if (link->state & DEV_RELEASE_PENDING) {
	del_timer(&link->release);
	link->state &= ~DEV_RELEASE_PENDING;
    }
    restore_flags(flags);

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
	tos_release((u_long)link);
	if (link->state & DEV_STALE_CONFIG) {
	    link->state |= DEV_STALE_LINK;
	    return;
	}
    }

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
	struct device *dev = link->priv;
	if (dev->priv)
	    kfree_s(dev->priv, sizeof(struct ei_device));
	kfree_s(dev, sizeof(struct tos_dev_t));
    }
    kfree_s(link, sizeof(struct dev_link_t));
    
} /* tos_detach */

/*======================================================================

    tos_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
    
======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn), args))!=0) goto cs_failed

static void tos_config(dev_link_t *link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    tos_dev_t *info;
    struct device *dev;
    int last_ret, last_fn; 
    int i, j;
    u_short buf[64];
    int manfid = 0, prodid = 0;

    handle = link->handle;
    info = link->priv;
    dev = &info->dev;

    DEBUG(0, "tos_config(0x%p)\n", link);

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.Attributes = 0;
    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    tuple.DesiredTuple = CISTPL_CONFIG;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
    link->state |= DEV_CONFIG;

    tuple.DesiredTuple = CISTPL_MANFID;
    tuple.Attributes = TUPLE_RETURN_COMMON;
    if ((CardServices(GetFirstTuple, handle, &tuple) == CS_SUCCESS) &&
 	(CardServices(GetTupleData, handle, &tuple) == CS_SUCCESS)) {
	manfid = le16_to_cpu(buf[0]);
	prodid = le16_to_cpu(buf[1]);
	DEBUG(0, "tos_cs: (manfid, prodid) = (%x, %x)\n", manfid, prodid);
    }

    /*
      Try allocating IO ports.  This tries a few fixed addresses.
      If you want, you can also read the card's config table to
      pick addresses -- see the serial driver for an example.
    */
    for (j = 0; j < 0x400; j += 0x20) {
	link->io.BasePort1 = j ^ 0x300;
	link->io.BasePort2 = (j ^ 0x300) + 0x10;
	i = CardServices(RequestIO, link->handle, &link->io);
	if (i == CS_SUCCESS) break;
    }

    if (i != CS_SUCCESS) {
	cs_error(handle, RequestIO, i);
	goto failed;
    }

    DEBUG(0, "OK CardServices/RequestIO %d\n", j);
	
    /*
      Now allocate an interrupt line.  Note that this does not
      actually assign a handler to the interrupt.
    */
    CS_CHECK(RequestIRQ, handle, &link->irq);

    /*
      This actually configures the PCMCIA socket -- setting up
      the I/O windows and the interrupt mapping.
    */
    CS_CHECK(RequestConfiguration, handle, &link->conf);

    /* The 8390 isn't at the base address -- the ASIC regs are there! */
    dev->base_addr = link->io.BasePort1+TOS_NIC_OFFSET;
    
    /* IRQ */
    dev->irq = link->irq.AssignedIRQ;
	
    /* misc */
    dev->tbusy = 0;

    /* OK, register it as network device */
    if (register_netdev(dev) != 0) {
	printk(KERN_NOTICE "tos_cs: register_netdev() failed\n");
	goto failed;
    }

    if (tos_setup_config_register(link) != 0) {
	printk(KERN_ERR "tos_cs: tos_setup_config_register() failed\n");
	goto failed;
    }

    /* yes, get a 32K of shared memory */
    if (setup_shmem_window(link, TOS_START_PG, TOS_STOP_PG, TOS_CM_OFFSET) != 0) {
	printk(KERN_ERR "tos_cs: setup_shmem_window() failed\n");
	goto failed;
    }
    
    tossetup(link);

    ei_status.name = hw_info->id;
    ei_status.word16 = 1;
    ei_status.reset_8390 = &tos_reset_8390;

    link->dev = &info->node;
    link->state &= ~DEV_CONFIG_PENDING;

    printk(KERN_NOTICE "tos device loaded\n");

    return;

#if 0
config_undo:
    unregister_netdev(dev);
    goto failed;
#endif

cs_failed:
    cs_error(link->handle, last_fn, last_ret);

failed:
    tos_release((u_long)link);
    return;

} /* tos_config */

static int tos_setup_config_register(dev_link_t *link)
{
    win_req_t req;
    memreq_t mem;
    u_char *virt;
    int i;

    DEBUG(0, "tos_config()\n");

    /* Allocate a 4K memory window */
    req.Attributes = WIN_DATA_WIDTH_8|WIN_MEMORY_TYPE_AM|WIN_ENABLE;
    req.Base = 0; req.Size = 0x1000;
    req.AccessSpeed = 0;
    link->win = (window_handle_t)link->handle;
    i = CardServices(RequestWindow, &link->win, &req);
    if (i != CS_SUCCESS) {
	cs_error(link->handle, RequestWindow, i);
	return i;
    }
    virt = ioremap(req.Base, 0x1000);
    mem.Page = 0;
    mem.CardOffset = 0;
    i = CardServices(MapMemPage, link->win, &mem);
    if (i != CS_SUCCESS) {
	cs_error(link->handle, MapMemPage, i);
	return i;
    }
    virt[TOS_KICKREG_OFFSET] = TOS_KICKREG_VAL;

    iounmap(virt);
    i = CardServices(ReleaseWindow, link->win);
    if (i != CS_SUCCESS) {
	cs_error(link->handle, ReleaseWindow, i);
	return i;
    }

    return 0;
}

/* setting up toshiba lan card --- specific operations */
static int tossetup(dev_link_t *link)
{
    int i;
    char *model_name;
    int reset_start_time;
    int reg2;
    int tos_ioaddr;
    struct tos_dev_t *info;
    struct device *dev;

    DEBUG(0, "tossetup()\n");

    info = link->priv;
    dev = &info->dev;
    
    tos_ioaddr  = link->io.BasePort1;

    /* Turn on the card. */
    outb(TOS_RESET|TOS_POWERON, tos_ioaddr);
    reset_start_time = jiffies;
    while ((inb(tos_ioaddr) & 0x4)) { /* wait for it to comes up */
	if (jiffies - reset_start_time > 8) {
	    break;
	}
    }
    outb(TOS_POWERON, tos_ioaddr);
    
    reg2 = 0;
    /* I think it's a type 4 because this driver is for cardmgr */
    switch (inb(tos_ioaddr + 8 + 6)) {
    case 0x11: model_name = "Type 1"; break;
    case 0x12: model_name = "Type 2"; reg2 = 0xf0; break;
    case 0x13: model_name = "Type 3"; reg2 = 0xf0; break;
    case 0x14: model_name = "Type 4"; break;
	default:   model_name = "unknown type";
    }
    
    printk(KERN_INFO "%s: Toshiba LAN I/F Card %s: port %#3x, irq %d,", 
	   dev->name, model_name, tos_ioaddr, dev->irq);
    printk(" shared memory at %#lx-%#lx,", dev->mem_start, dev->mem_end-1);

    printk(" hw_addr ");
    for (i = 0; i < ETHER_ADDR_LEN; i++)
	printk("%02X%s",
	       (dev->dev_addr[i] = inb(tos_ioaddr+TOS_EADDR_OFFSET+i)),
	       ((i<5) ? ":" : "\n"));
    printk(KERN_NOTICE "Driver Version: %s\n", version);

    return 0;
}

/*======================================================================

    After a card is removed, tos_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void tos_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    tos_dev_t *info = link->priv;

    DEBUG(0, "tos_release(0x%p)\n", link);

    /*
       If the device is currently in use, we won't release until it
       is actually closed.
    */
    if (link->open) {
	DEBUG(1, "tos_cs: release postponed, '%s' still open\n", 
	      info->node.dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    /* Don't bother checking to see if these succeed or not */
    iounmap(info->base);
    CardServices(ReleaseWindow, link->win);
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);

    if (link->dev)
	unregister_netdev(&info->dev);
    link->dev = NULL;

#if	CS_RELEASE_CODE < 0x2911
    if (info->dev.irq != 0)
	irq2dev_map[info->dev.irq] = NULL;
#endif

    link->state &= ~(DEV_CONFIG | DEV_RELEASE_PENDING);
    
    if (link->state & DEV_STALE_LINK)
	tos_detach(link);
    
} /* tos_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.

    When a CARD_REMOVAL event is received, we immediately set a flag
    to block future accesses to this device.  All the functions that
    actually access the device should check this flag to make sure
    the card is still present.
    
======================================================================*/

static int tos_event(event_t event, int priority,
			  event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    tos_dev_t *info = link->priv;

    DEBUG(2, "tos_event(0x%06x)\n", event);
    
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    info->dev.tbusy = 1; info->dev.start = 0;
	    link->release.expires = RUN_AT(HZ/20);
	    link->state |= DEV_RELEASE_PENDING;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT;
	tos_config(link);
	break;
    case CS_EVENT_PM_SUSPEND:
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
        if (link->state & DEV_CONFIG) {
	    if (link->open) {
		info->dev.tbusy = 1; info->dev.start = 0;
	    }
	    CardServices(ReleaseConfiguration, link->handle);
	}
	break;
    case CS_EVENT_PM_RESUME:
	link->state &= ~DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_CARD_RESET:
	printk(KERN_NOTICE "CARD_RESET\n");
	if (link->state & DEV_CONFIG) {
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	    if (link->open) {
		tos_reset_8390(&info->dev);
		NS8390_init(&info->dev, 1);
		info->dev.tbusy = 0; info->dev.start = 1;
	    }
	    do {
		/* XXX kick him */
		int i;
		modwin_t mod;
		memreq_t mem;

		mod.Attributes = WIN_DATA_WIDTH_8|WIN_MEMORY_TYPE_AM|WIN_ENABLE;
		i = CardServices(ModifyWindow, link->win, &mod);
		if (i != 0) {
		    cs_error(link->handle, ModifyWindow, i);
		    break;
		}
	    
		/* First, use the window to get the hardware address */
		mem.CardOffset = 0; mem.Page = 0;
		i = CardServices(MapMemPage, link->win, &mem);
		if (i != 0) {
		    cs_error(link->handle, MapMemPage, i);
		    break;
		}
	       
		/* The card needs the byte. But Why? */
		((caddr_t)(info->dev.mem_start))[TOS_KICKREG_OFFSET] = TOS_KICKREG_VAL; 

		mod.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM|WIN_ENABLE;
		i = CardServices(ModifyWindow, link->win, &mod);
		if (i != 0) {
		    cs_error(link->handle, ModifyWindow, i);
		    break;
		}
		mem.CardOffset = TOS_START_PG<<8;
		i = CardServices(MapMemPage, link->win, &mem);
		if (i != 0) {
		    cs_error(link->handle, MapMemPage, i);
		    break;
		}
	    } while (0);
	    /* what can I do when error occurs? */
	}
	break;
    }
    return 0;
} /* tos_event */


/*====================================================================*/

static int tos_open(struct device *dev)
{
    dev_link_t *link;
    int tos_ioaddr = DS_BASE - TOS_NIC_OFFSET; /* TOS_CMDREG */
    
    DEBUG(2, "tos_open('%s')\n", dev->name);
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if (!DEV_OK(link))
	return -ENODEV;
    
    link->open++;
    MOD_INC_USE_COUNT;
    
    /* Map in the shared memory. */
    outb(((dev->mem_start&0xe000)>>8) +
	 ((((ei_status.stop_page - TOS_START_PG)*256)&0x1e000) >> 13),
	 tos_ioaddr+TOS_CMDREG1);
    outb((dev->mem_start>>16) | ei_status.reg0, tos_ioaddr+TOS_CMDREG2);
    outb(TOS_POWERON|TOS_MEMENB, tos_ioaddr); /* WD_CMDREG */
    
    return ei_open(dev);
} /* tos_open */

/*====================================================================*/

static int tos_close(struct device *dev)
{
    dev_link_t *link;
    
    DEBUG(2, "tos_close('%s')\n", dev->name);
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if (link == NULL)
	return -ENODEV;
    FREE_IRQ(dev->irq, dev);

    link->open--; dev->start = 0;
    if (link->state & DEV_STALE_CONFIG) {
	link->release.expires = RUN_AT(HZ/20);
	link->state |= DEV_RELEASE_PENDING;
	add_timer(&link->release);
    }

    MOD_DEC_USE_COUNT;
    
    return 0;
} /* tos_close */


/*======================================================================
  
  Hard reset the card.  This used to pause for the same period that
  a 8390 reset command required, but that shouldn't be necessary.
  
  ======================================================================*/

static void tos_reset_8390(struct device *dev)
{
    
    int tos_cmd_port = DS_BASE - TOS_NIC_OFFSET; /* TOS_CMDREG */
    
    ei_status.txing = 0;
    outb(TOS_POWERON|TOS_RESET, tos_cmd_port);
    
    /* Set up the ASIC registers, just in case something changed them. */
    outb(TOS_POWERON|TOS_MEMENB, tos_cmd_port);
    
    /* hmmm... I don't know whether reset suceeded or not */ 
    if (ei_debug > 1) printk(KERN_DEBUG "reset done\n");
    return;
    
} /* tos_reset_8390 */

/*====================================================================*/

void tos_interrupt(int reg)
{
    printk(KERN_NOTICE "tos_cs: interrupt\n");
} /* tos_interrupt */

/*====================================================================*/

static void copyin(unsigned char *dest, unsigned char *src, int c)
{
    unsigned short *d = (unsigned short *) dest;
    unsigned short *s = (unsigned short *) src;
    int odd;

    if (c <= 0)
	return;
    odd = (c & 01); c >>= 1;

    if (c) {
	do { *d++ = readw_ns(s++); } while (--c);
    }
    /* get last byte by fetching a word and masking */
    if (odd)
	*((unsigned char *)d) = readw(s) & 0xff;
}

static void copyout(unsigned char *dest, const unsigned char *src, int c)
{
    volatile unsigned short *d = (unsigned short *) dest;
    unsigned short *s = (unsigned short *) src;
    int odd;

    if (c <= 0)
	return;
    odd = (c & 01); c >>= 1;

    if (c) {
	do { writew_ns(*s++, d++); } while (--c);
    }
    /* copy last byte doing a read-modify-write */
    if (odd)
	writew((readw(d) & 0xff00) | *(u_char *)s, d);
}

/*====================================================================*/

static void shmem_get_8390_hdr(struct device *dev,
			       struct e8390_pkt_hdr *hdr,
			       int ring_page)
{
    void *xfer_start = (void *)(dev->rmem_start + (ring_page << 8)
				- (ei_status.rx_start_page << 8));
    
    copyin((void *)hdr, xfer_start, sizeof(struct e8390_pkt_hdr));
}

/*====================================================================*/

static void shmem_block_input(struct device *dev, int count,
			      struct sk_buff *skb, int ring_offset)
{
    void *xfer_start = (void *)(dev->rmem_start + ring_offset
				- (ei_status.rx_start_page << 8));
    char *buf = skb->data;
    
    if (xfer_start + count > (void *)dev->rmem_end) {
	/* We must wrap the input move. */
	int semi_count = (void*)dev->rmem_end - xfer_start;
	copyin(buf, xfer_start, semi_count);
	buf += semi_count;
	ring_offset = ei_status.rx_start_page << 8;
	xfer_start = (void *)dev->rmem_start;
	count -= semi_count;
    }
    copyin(buf, xfer_start, count);
}

/*====================================================================*/

static void shmem_block_output(struct device *dev, int count,
			       const unsigned char *buf,
			       const int start_page)
{
    void *shmem = (void *)dev->mem_start + (start_page << 8);
    shmem -= ei_status.tx_start_page << 8;

    if (ei_debug > 4)
	printk(KERN_DEBUG "[bo=%d @ %x]\n", count, start_page);

    copyout(shmem, buf, count);
}

/*====================================================================*/

static int setup_shmem_window(dev_link_t *link, int start_pg,
			      int stop_pg, int cm_offset)
{
    win_req_t req;
    memreq_t mem;
    int offset = 0, last_ret, last_fn;
    struct device *dev = link->priv;
    int window_size;
    tos_dev_t *info = link->priv;

    window_size = (stop_pg - start_pg) << 8;
    if (window_size > 32 * 1024)
	window_size = 32 * 1024;

    /* Make sure it's a power of two.  */
    while ((window_size & (window_size - 1)) != 0)
	window_size += window_size & ~(window_size - 1);

    /* Allocate a memory window */
    req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM|WIN_ENABLE;
    req.Attributes |= WIN_USE_WAIT;
    req.Base = 0; req.Size = window_size;
    req.AccessSpeed = mem_speed;
    link->win = (window_handle_t)link->handle;
    CS_CHECK(RequestWindow, &link->win, &req);

    mem.CardOffset = (start_pg << 8) + cm_offset;
    offset = mem.CardOffset % window_size;
    mem.CardOffset -= offset;
    mem.Page = 0;
    CS_CHECK(MapMemPage, link->win, &mem);

    info->base = ioremap(req.Base, window_size);
    dev->mem_start = (u_long)info->base + offset;
    dev->rmem_start = dev->mem_start + (TX_PAGES<<8);
    dev->mem_end = dev->rmem_end = (u_long)info->base + req.Size;

    ei_status.tx_start_page = start_pg;
    ei_status.rx_start_page = start_pg + TX_PAGES;
    ei_status.stop_page = start_pg + ((req.Size - offset) >> 8);

    /* set up block i/o functions */
    ei_status.get_8390_hdr = &shmem_get_8390_hdr;
    ei_status.block_input = &shmem_block_input;
    ei_status.block_output = &shmem_block_output;

    return 0;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
    return 1;
}

static int set_config(struct device *dev, struct ifmap *map)
{
    printk(KERN_INFO "%s: sorry. support 10BASE-T only\n", dev->name);
    return 0;
}

/*====================================================================*/

int init_module(void)
{
    servinfo_t serv;
    DEBUG(0, "%s\n", version);
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_NOTICE "tos_cs: Card Services release "
	       "does not match!\n");
	return -1;
    }
    register_pcmcia_driver(&dev_info, &tos_attach, &tos_detach);
    return 0;
}

void cleanup_module(void)
{
    DEBUG(0, "tos_cs: unloading\n");
    unregister_pcmcia_driver(&dev_info);
    while (dev_list != NULL)
	tos_detach(dev_list);
}

/*
 * Local Variables:
 * c-basic-offset: 4
 * c-file-offsets: ((label . 0) (case-label . 0) (statement-case-intro . 0))
 * End:
 */
